from PyQt6.QtCore import Qt, QModelIndex, pyqtSignal
from PyQt6.QtGui import QStandardItemModel, QStandardItem, QFont, QAction
from PyQt6.QtWidgets import (
    QApplication,
    QMainWindow,
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QSplitter,
    QTreeView,
    QPlainTextEdit,
    QPushButton,
    QLabel,
    QFileDialog,
    QMessageBox,
    QGroupBox,
    QStatusBar,
)



from .imports import *


# ── helpers ──────────────────────────────────────────────────────────────
def _replace_log(self, text: str):
    try:
        self.log_view.clear()
        self.log_view.insertPlainText(text)
    except Exception as e:
        print(f"{e}")

def _parse_item(self, info: str):
    try:
        parts = info.rsplit(":", 2)
        if len(parts) == 3:
            path, line, col = parts[0], parts[1], parts[2]
        else:
            path, line, col = parts[0], parts[1], "1"
        return path, int(line), int(col)
    except Exception as e:
        print(f"{e}")

def _extract_errors_for_file(self, combined_text: str, abs_path: str, project_root: str) -> str:
    try:
        text = combined_text or ""
        if not text:
            return ""
        try:
            rel = os.path.relpath(abs_path, project_root) if (project_root and abs_path.startswith(project_root)) else os.path.basename(abs_path)
        except Exception:
            rel = os.path.basename(abs_path)
        rel_alt = rel.replace("\\", "/")
        abs_alt = abs_path.replace("\\", "/")
        base = os.path.basename(abs_alt)
        lines = text.splitlines()
        blocks = []
        for i, ln in enumerate(lines):
            if (abs_alt in ln) or (rel_alt in ln) or (("src/" + base) in ln):
                start = max(0, i - 3)
                end = min(len(lines), i + 6)
                block = "\n".join(lines[start:end])
                blocks.append(f"\n— context @ log line {i+1} —\n{block}\n")
        return "\n".join(blocks).strip()
    except Exception as e:
        print(f"{e}")

def create_radio_group(self, labels, default_index=0, slot=None):
    group = QButtonGroup(self)
    buttons = []
    for i, label in enumerate(labels):
        rb = QRadioButton(label)
        if i == default_index:
            rb.setChecked(True)
        group.addButton(rb)
        buttons.append(rb)
        if slot:
            rb.toggled.connect(slot)
    return group, buttons

_PREFERRED_EXT_ORDER = [".tsx",".ts",".jsx",".js",".mjs",".cjs",".tsx",".ts",".jsx",".js",".css",".scss",".less"]
def resolve_alt_ext(self,path: str, project_root: str) -> str:
    """
    If 'path' doesn't exist, try swapping extensions using a common React stack order.
    Also try replacing absolute project-root prefix with relative 'src/' and vice-versa.
    """
    try_paths = [path]
    base, ext = os.path.splitext(path)
    for e in _PREFERRED_EXT_ORDER:
        try_paths.append(base + e)

    # also try joining with project_root, and under src/
    if project_root and not path.startswith(project_root):
        rel = path.lstrip("./")
        try_paths.extend([
            os.path.join(project_root, rel),
            os.path.join(project_root, "src", os.path.basename(base)) + ext,
        ])
        for e in _PREFERRED_EXT_ORDER:
            try_paths.append(os.path.join(project_root, "src", os.path.basename(base)) + e)

    for candidate in try_paths:
        if os.path.isfile(candidate):
            return candidate
    return path  # fallback

def _editor_save_current(self):
    if not self.current_file_path:
        QMessageBox.information(self, "Save", "No file loaded.")
        return
    try:
        text = self.editor.toPlainText()
        with open(self.current_file_path, "w", encoding="utf-8") as f:
            f.write(text)
        self.original_text = text
        self.append_log(f"[editor] saved: {self.current_file_path}\n")
    except Exception as e:
        QMessageBox.critical(self, "Save Error", str(e))
        self.append_log(f"[editor] save error: {e}\n")

def _editor_revert_current(self):
    if not self.current_file_path:
        return
    self.editor.setPlainText(self.original_text or "")
    self.append_log(f"[editor] reverted: {self.current_file_path}\n")
def _editor_show_ranges(self, path: str, ranges: list[dict], center: tuple[int,int]|None=None):
    """
    ranges: [{line:int, col:int, message:str, code:str}]
    """
    try:
        if not hasattr(self, "editor") or not self.editor:
            return
        if hasattr(self.editor, "set_document"):
            self.editor.set_document(path)  # ensure the right file is loaded
        if hasattr(self.editor, "highlight_ranges"):
            # your editor can decide color by severity/ code; we pass metadata
            self.editor.highlight_ranges(ranges, center=center)
        elif hasattr(self.editor, "highlight_range"):  # legacy single-call
            for r in ranges:
                self.editor.highlight_range(r["line"], r["col"])
            if center:
                self.editor.center_on_line(center[0])
    except Exception:
        pass



from .imports import *
def start_work(self):
    try:
        self.run_btn.setEnabled(False)
        user = self.user_in.text().strip() or 'solcatcher'   # <- swap order (yours hard-coded the default)
        path = self.path_in.text().strip()
        if not path or not os.path.isdir(path):
            QMessageBox.critical(self, "Error", "Invalid project path.")
            self.run_btn.setEnabled(True)
            return

        # Clear old UI bits
        self.errors_tree.clear()
        self.warnings_tree.clear()
        
        # Kick off non-blocking build
        self._run_build_qprocess(path)
        
    except Exception:
        self.append_log("start_work error:\n" + traceback.format_exc() + "\n")
        self.run_btn.setEnabled(True)

def clear_ui(self):
    self.log_view.clear()
    self.errors_tree.clear()
    self.warnings_tree.clear()
    self.last_output = ""
    self.last_errors_only = ""
    self.last_warnings_only = ""



from PyQt6.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QTextEdit, QListWidget,
    QTabWidget, QLineEdit, QLabel, QFormLayout, QScrollArea, QListWidgetItem
    
)
from PyQt6.QtWidgets import QComboBox
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QDesktopServices
from PyQt6.QtCore import QUrl



from .imports import *
from .green_screen_delimiter import *
# From functions.py
def run_all(input_path: str, weak: float, strong: float, out_dir_name="new_pics"):
    base_dir = os.path.dirname(input_path)
    out_dir = os.path.join(base_dir, out_dir_name)
    os.makedirs(out_dir, exist_ok=True)

    all_path = os.path.join(out_dir, "all_green.png")
    blur_path = os.path.join(out_dir, "green_screen_blur.png")
    blur_path_thresh = os.path.join(out_dir, "green_screen_blur.png")
    detect_all_green(input_path, all_path, strong_thresh=strong, weak_thresh=weak)
    detect_greenscreen_blur(input_path, blur_path)
    detect_greenscreen_blur_thresh(input_path, blur_path_thresh, strong_thresh=strong, weak_thresh=weak)

    return input_path, all_path, blur_path,blur_path_thresh

class LabeledImage(QVBoxLayout):
    """Small helper layout: a bold label and a scaled image view."""
    def __init__(self, title: str):
        super().__init__()
        self.title = QLabel(f"<b>{title}</b>")
        self.title.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.view = QLabel("Select an image to begin")
        self.view.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.view.setFrameShape(QFrame.Shape.StyledPanel)
        self.view.setMinimumSize(320, 240)

        self.addWidget(self.title)
        self.addWidget(self.view)

    def set_pixmap(self, pix: QPixmap):
        self.view.setPixmap(pix)



from abstract_gui.QT6 import *
from abstract_gui.QT6.utils import *
from abstract_react import *
from abstract_utilities import *
from .ext_funcs import *
from .init_imports import *
from .constants import *
from .init_imports import *
from .pyqt_imports import *



from ..src.constants import *
import webbrowser
# ── worker ────────────────────────────────────────────────────────────────────

def request_captcha() -> tuple[str, str]:
    body = requests.post(f'{API_BASE}/requestRecaptcha').json()
    print("[captcha]", body)
    return body['challenge'], body['captcha_url']


def login(email: str, password: str, challenge: str, captcha_answer: str) -> str:
    body = requests.post(f'{API_BASE}/login', json={
        'username': email,
        'password': password,
        're_captcha_challenge': challenge,
        're_captcha_response': captcha_answer,
    }).json()
    print("[login]", body)
    if body.get('status') != 'success':
        return False
    return body['auth_token']


def get_api_url(auth_token: str, file_id: str) -> str:
    body = requests.post(f'{API_BASE}/getUrl', json={
        'auth_token': auth_token,
        'file_id': file_id,
    }).json()
    print("[getUrl]", body)
    if body.get('status') != 'success':
        return False
    return body['url']


def download(url: str,name=None) -> Path:
    OUT_DIR.mkdir(parents=True, exist_ok=True)
    name = name or url.rsplit('/', 1)[-1].split('?')[0]
    dest = OUT_DIR / name
    if not os.path.isfile(str(dest)):
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(dest, 'wb') as f:
                for chunk in r.iter_content(8192):
                    f.write(chunk)
    return dest

def get_token(token=None):
    if not token:
        challenge, captcha_url = request_captcha()
        print(f"\nOpen this URL and read the captcha text:\n{captcha_url}\n")
        webbrowser.open(captcha_url)
        answer = input("Captcha answer: ").strip()
        token = login(userName, passWord, challenge, answer)
        print(token)
    return token



from PyQt6 import QtCore
from PyQt6.QtWidgets import QToolButton,QHBoxLayout,QSpinBox,QGridLayout,QWidget,QCheckBox,QLineEdit,QPushButton,QLabel
from PyQt6.QtCore import QSignalBlocker
from ..imports import *

from PyQt6.QtCore import QObject, pyqtSignal
from .results import *
from .states import *
from .open_file_funcs import *
from abstract_utilities.file_utils import (derive_file_defaults)

from .visibility import *
logger = get_logFile(__name__)
_read_state = read_state
_write_state = write_state
def normalize_filter_value(v):
    if v is None or v is False:
        return ""
    if isinstance(v, (set, list, tuple)):
        return ",".join(sorted(str(x) for x in v))
    return str(v)

class SharedStateBus(QObject):
    stateBroadcast = pyqtSignal(object, dict)  # (sender, state)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._snap: dict = {}

    def snapshot(self) -> dict:
        return dict(self._snap)

    def push(self, sender, state: dict):
        self._snap = dict(state)
        self.stateBroadcast.emit(sender, self.snapshot())
def getContentComponentDefaults(
        host,
        *args,
        primary_btn=None,
        secondary_btn=None,
        trinary_btn=None,
        **kwargs
    ):
    DEFAULT_SECONDARY_BUTTON=("blank", None) 
    DEFAULT_PRIMARY_BUTTON = ("Run", None)
    DEFAULT_TRINARY_BUTTON = ("open all hits", host.open_all_hits)
    host.primary_btn=primary_btn or DEFAULT_PRIMARY_BUTTON
    host.secondary_btn=secondary_btn or  DEFAULT_SECONDARY_BUTTON
    host.trinary_btn = trinary_btn or  DEFAULT_TRINARY_BUTTON
    file_defaults = derive_file_defaults(**kwargs)
    logger.info(file_defaults)
    host.allowed_exts_in   = QLineEdit(normalize_filter_value(file_defaults.get('allowed_exts')))
    host.exclude_exts_in = host.unallowed_exts_in =QLineEdit(normalize_filter_value(file_defaults.get('exclude_exts')))
    host.allowed_types_in  = QLineEdit(normalize_filter_value(file_defaults.get('allowed_types')))
    host.exclude_types_in  = QLineEdit(normalize_filter_value(file_defaults.get('exclude_types')))
    host.allowed_dirs_in   = QLineEdit(normalize_filter_value(file_defaults.get('allowed_dirs')))
    host.exclude_dirs_in   = QLineEdit(normalize_filter_value(file_defaults.get('exclude_dirs')))
    host.allowed_patterns_in = QLineEdit(normalize_filter_value(file_defaults.get('allowed_patterns')))
    host.exclude_patterns_in = QLineEdit(normalize_filter_value(file_defaults.get('exclude_patterns')))

    host.chk_add        = QCheckBox("Add");          host.chk_add.setChecked(False)
    host.chk_recursive  = QCheckBox("Recursive");    host.chk_recursive.setChecked(True)
    host.chk_total      = QCheckBox("Require ALL strings (total_strings)")
    host.chk_total.setChecked(False)
    host.chk_parse      = QCheckBox("parse_lines");  host.chk_parse.setChecked(False)
    host.chk_getlines   = QCheckBox("get_lines");    host.chk_getlines.setChecked(True)

    host.spec_spin = QSpinBox(); host.spec_spin.setRange(0, 999999); host.spec_spin.setValue(0)
    global_default_filters = dict(
                **file_defaults,
                add=host.chk_add 
            )
    return global_default_filters
def createLayoutForm(host):
    r = 0
    host._input_grid.addWidget(QLabel("Directory"), r, 0); host._input_grid.addWidget(host.dir_in, r, 1); host._input_grid.addWidget(host.btn_browse, r, 2); r+=1
    host._input_grid.addWidget(QLabel("Strings"), r, 0); host._input_grid.addWidget(host.strings_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Allowed Exts"), r, 0); host._input_grid.addWidget(host.allowed_exts_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Exclude Exts"), r, 0); host._input_grid.addWidget(host.exclude_exts_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Allowed Types"), r, 0); host._input_grid.addWidget(host.allowed_types_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Exclude Types"), r, 0); host._input_grid.addWidget(host.exclude_types_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Allowed Dirs"), r, 0); host._input_grid.addWidget(host.allowed_dirs_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Exclude Dirs"), r, 0); host._input_grid.addWidget(host.exclude_dirs_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Allowed Patterns"), r, 0); host._input_grid.addWidget(host.allowed_patterns_in, r, 1, 1, 2); r+=1
    host._input_grid.addWidget(QLabel("Exclude Patterns"), r, 0); host._input_grid.addWidget(host.exclude_patterns_in, r, 1, 1, 2); r+=1

    flags = QHBoxLayout()
    for w in (host.chk_recursive, host.chk_total, host.chk_parse, host.chk_getlines, host.chk_add):
        flags.addWidget(w)
    host._input_grid.addLayout(flags, r, 0, 1, 3); r+=1

    sp = QHBoxLayout()
    sp.addWidget(QLabel("spec_line (0=off):")); sp.addWidget(host.spec_spin); sp.addStretch(1)
    host._input_grid.addLayout(sp, r, 0, 1, 3); r+=1

    host.layout().addWidget(host._input_container)
    # CTA row
    cta = QHBoxLayout()
    if host.primary_btn[1]:
        host.btn_run = QPushButton(host.primary_btn[0]); host.btn_run.clicked.connect(host.primary_btn[1])
        cta.addWidget(host.btn_run)
    cta.addStretch(1)
    if host.secondary_btn[1]:
        host.btn_secondary = QPushButton(host.secondary_btn[0]); host.btn_secondary.clicked.connect(host.secondary_btn[1])
        cta.addWidget(host.btn_secondary)
    if host.trinary_btn[1]:
        host.btn_third = QPushButton(host.trinary_btn[0]); host.btn_third.clicked.connect(host.trinary_btn[1])
        cta.addWidget(host.btn_third)
    host.layout().addLayout(cta)
def getSharedBus(
        host,
        bus: SharedStateBus,
        auto_unlink_on_init_if_diff=True,
        global_default_filters=None
    ):

    host._bus = bus
    host._applying_remote = False

    # --- propagate local → bus when linked
    def maybe_broadcast(*_):
        if not host.link_btn.isChecked() or host._applying_remote:
            return
        bus.push(host, _read_state(host))

    # connect all widgets to broadcast
    for w in (host.dir_in, host.strings_in,
              host.allowed_exts_in, host.exclude_exts_in,
              host.allowed_types_in, host.exclude_types_in,
              host.allowed_dirs_in, host.exclude_dirs_in,
              host.allowed_patterns_in, host.exclude_patterns_in):
        w.textEdited.connect(maybe_broadcast)

    for cb in (host.chk_add, host.chk_recursive, host.chk_total,
               host.chk_parse, host.chk_getlines):
        cb.toggled.connect(maybe_broadcast)

    host.spec_spin.valueChanged.connect(maybe_broadcast)

    # --- bus → host
    def apply_shared(sender, state):
        if sender is host or not host.link_btn.isChecked():
            return
        host._applying_remote = True
        _write_state(host, state)
        host._applying_remote = False

    bus.stateBroadcast.connect(apply_shared)

    # --- INITIAL SYNC
    current = _read_state(host)
    snap = bus.snapshot()

    if snap:
        # Already have shared state → sync with that
        host._applying_remote = True
        _write_state(host, snap)
        host._applying_remote = False
        return

    # First tab to attach → seed the bus
    bus.push(host, current)


def add_result_item(list_widget, file_path: str, line: int | None):
    text = f"{file_path}:{line}" if line is not None else file_path
    it = QListWidgetItem(text)
    it.setData(Qt.ItemDataRole.UserRole, {"file_path": file_path, "line": line})
    list_widget.addItem(it)
def setupLayout(host, grid=None):
    grid = grid or QGridLayout()
    host._input_grid = grid
    form_container = QWidget()
    form_container.setLayout(host._input_grid)
    host._input_container = form_container
    host.browse_dir = browse_dir
    host.make_params = make_params

    # bind editor openers so `self` is host
    host.open_one = MethodType(open_one, host)
    host.open_all_hits = MethodType(open_all_hits, host)

    # do not create or insert a list here; tabs manage their own `self.list`
    if hasattr(host, "list"):
        try:
            host.list.itemDoubleClicked.disconnect()
        except Exception:
            pass
        host.list.itemDoubleClicked.connect(host.open_one)
def install_common_inputs(host, grid: QGridLayout, *,
                          bus: SharedStateBus,
                          primary_btn=None,
                          secondary_btn=None,
                          trinary_btn=None,
                          default_dir_in=None,
                          default_allowed_exts_in=None,
                          default_exclude_exts_in=None,
                          default_allowed_types_in=None,
                          default_exclude_types_in=None,
                          default_allowed_dirs_in=None,
                          default_exclude_dirs_in=None,
                          default_allowed_patterns_in=None,
                          default_exclude_patterns_in=None,
                          # NEW:
                          auto_unlink_on_init_if_diff=True,
                          global_default_filters: dict | None = None):

    """
    Adds the shared controls to `host` and wires them to `bus`.
    Expects `host` to already have a QVBoxLayout() as its layout().
    """

    setupLayout(host,grid=grid)
    # --- widgets (your exact fields) -------------------------------------
    host.dir_in = QLineEdit(default_dir_in or os.getcwd())
    host.btn_browse = QPushButton("Browse…")
    host.btn_browse.clicked.connect(lambda: browse_dir(host))
    host._vis = visibilityMgr(host, settings_app="common_inputs", animate_default=True)
    host.strings_in = QLineEdit("")
    host.strings_in.setPlaceholderText("comma,separated,strings")
    
    global_default_filters = getContentComponentDefaults(
        host,
        primary_btn=primary_btn,
        secondary_btn=secondary_btn,
        trinary_btn=trinary_btn,
        allowed_exts=default_allowed_exts_in,
        exclude_exts=default_exclude_exts_in,
        allowed_types=default_allowed_types_in,
        exclude_types=default_exclude_types_in,
        allowed_dirs=default_allowed_dirs_in,
        exclude_dirs=default_exclude_dirs_in,
        allowed_patterns=default_allowed_patterns_in,
        exclude_patterns=default_exclude_patterns_in
    )
    logger.info(global_default_filters)
    # toolbar row you already have
    link_row = QHBoxLayout()
    host.link_btn = QToolButton(); host.link_btn.setCheckable(True); host.link_btn.setChecked(True)
    host.link_btn.setText("🔗 Linked"); link_row.addWidget(host.link_btn)
    link_row.addStretch(1)
    host.layout().addLayout(link_row)

    # NEW: visibility manager
    if not hasattr(host, "_vis"):
        host._vis = VisibilityMgr(host, settings_app="FinderConsole", animate_default=True)

    # Register the form as a collapsible section, auto-create a toggle button and place it in link_row
    host._vis.collapse_btn = host._vis.register(
        name="filters",
        container=host._input_container,
        button_host_layout=link_row,          # puts the button next to "Linked" control
        start_visible=True,
        animate=True,                         # slide animation
        shortcut="Ctrl+`",                    # toggle with a hotkey
        button_text_open="−",
        button_text_closed="+",
        persist=True
    )
    # --- lay out form -----------------------------------------------------
    createLayoutForm(host)
    getSharedBus(host,bus=bus,global_default_filters=global_default_filters,auto_unlink_on_init_if_diff=auto_unlink_on_init_if_diff,)
    



from .imports import *
# ───────────────── helpers to read tree state ─────────────────

def _iter_tree_rows(self):
    """Yield (path, apply_checked, overwrite_checked, item) for each row."""
    for i in range(self.files_list.topLevelItemCount()):
        it = self.files_list.topLevelItem(i)
        path = it.data(0, Qt.ItemDataRole.UserRole) or it.text(0)
        apply_checked    = (it.checkState(1) == Qt.CheckState.Checked)
        overwrite_checked= (it.checkState(2) == Qt.CheckState.Checked)
        yield path, apply_checked, overwrite_checked, it

def _gather_checked(self):
    """Return list of dicts for rows the user wants to apply to."""
    targets = []
    for path, apply_ck, over_ck, _ in _iter_tree_rows(self):
        if apply_ck and path:
            targets.append({"path": path, "overwrite": over_ck})
    return targets

# ───────────── logging to log pane (if present) ─────────────
def get_files(self) -> list[str]:
    params = make_params(self)
    dirs, files = get_files_and_dirs(**params)
    return files
def append_log(self, text: str):
    if not hasattr(self, "log") or self.log is None:
        return
    c = self.log.textCursor()
    c.movePosition(c.MoveOperation.End)
    self.log.setTextCursor(c)
    self.log.insertPlainText(text)
    self.log.ensureCursorVisible()

# ───────────── small helpers for tests / diagnostics ────────

def get_all_files(self) -> list[str]:
    try:
        files = get_files(self)
    except Exception as e:
        QMessageBox.critical(self, "Error", str(e))
        set_status(self, f"Error: {e}", "error")
        return []
    if hasattr(self, "output"):
        self.output.insertPlainText(f"files = {files}\n")
    return files

def get_hunks(self, diff_text: str) -> list[Hunk]:
    hunks = parse_unified_diff(diff_text)
    if not hunks:
        QMessageBox.warning(self, "Warning", "No valid hunks found in diff.")
        set_status(self, "No valid hunks found.", "warn")
        return []
    if hasattr(self, "output"):
        self.output.insertPlainText(f"hunks = {hunks}\n")
    return hunks

def get_all_subs(self, hunks: list[Hunk]) -> list[str]:
    flat: list[str] = []
    for h in hunks:
        flat.extend(h.subs)
    if hasattr(self, "output"):
        self.output.insertPlainText(f"subs(flat) = {flat}\n")
    return flat

def find_matches_for_hunks(files: list[str], hunks: list[Hunk]) -> tuple[list[str], list[dict]]:
    all_files: set[str] = set()
    all_found: list[dict] = []
    for h in hunks:
        if not h.subs:
            continue
        nu_files, found_paths = getPaths(files, h.subs)
        all_files.update(nu_files)
        for fp in found_paths:
            fp["hunk"] = h
        all_found.extend(found_paths)
    return sorted(all_files), all_found

def get_test_diff(self) -> str:
    diff_text = """\
-def browse_dir(self):
-    d = QFileDialog.getExistingDirectory(self, "Choose directory", self.dir_in.text() or os.getcwd())
-    if d:
-        self.dir_in.setText(d)
+def browse_dir(self):
+    d = QFileDialog.getExistingDirectory(self, "Choose directory", self.dir_in.text() or os.getcwd())
+    if d:
+        self.dir_in.setText(d)
"""
    diff_text = diff_text.strip("\n")
    if hasattr(self, "output"):
        self.output.insertPlainText(f"diff_text = {diff_text}\n")
    return diff_text

def get_nufiles(self, files: list[str], subs: list[str] | str):
    nu_files, found_paths = getPaths(files, subs)
    if hasattr(self, "output"):
        self.output.insertPlainText(f"nu_files = {nu_files}\n")
        self.output.insertPlainText(f"found_paths = {found_paths}\n")
    return nu_files, found_paths

def output_test(self, *_):
    files = get_all_files(self)
    diff_text = get_test_diff(self)
    hunks = get_hunks(self, diff_text)
    subs = get_all_subs(self, hunks)
    get_nufiles(self, files, subs)

# ───────────────────── Core preview/apply ───────────────────

def _ask_user_to_pick_file(self, files: List[str], title: str = "Pick a file to preview") -> str | None:
    if not files:
        return None
    if len(files) == 1:
        return files[0]
    dlg = QFileDialog(self, title, os.path.dirname(files[0]) if files else os.getcwd())
    dlg.setFileMode(QFileDialog.FileMode.ExistingFile)
    dlg.setNameFilter("All files (*)")
    if dlg.exec():
        sel = dlg.selectedFiles()
        return sel[0] if sel else None
    return None



def apply_custom_diff(original_lines: List[str], diff_lines: List[str]) -> str:
    """
    Apply simplified unified diff: replace each exact multi-line 'subs' block with 'adds'.
    """
    if diff_lines and "/" in diff_lines[0]:
        diff_lines = diff_lines[1:]
    hunks = parse_unified_diff("\n".join(diff_lines))
    replacements = []
    og_content = "\n".join(original_lines)

    for h in hunks:
        if not h.subs:
            continue
        tot_subs = "\n".join(h.subs)
        for m in re.finditer(re.escape(tot_subs), og_content):
            start_byte = m.start()
            start_line = og_content[:start_byte].count("\n")
            if original_lines[start_line:start_line + len(h.subs)] == h.subs:
                replacements.append({"start": start_line, "end": start_line + len(h.subs), "adds": h.adds[:]})

    replacements.sort(key=lambda r: r["start"])
    for i in range(1, len(replacements)):
        if replacements[i-1]["end"] > replacements[i]["start"]:
            raise ValueError("Overlapping hunks detected.")

    out = original_lines[:]
    for r in reversed(replacements):
        out = out[:r["start"]] + r["adds"] + out[r["end"]:]
    return "\n".join(out)
def save_all_checked(self, make_backup: bool = True, backup_ext: str = ".bak"):
    """
    Save the current diff preview into ALL rows where:
      - Apply is checked, and
      - Overwrite determines whether we replace the file or write *.new
    """
    diff_text = self.diff_text.toPlainText().strip()
    if not diff_text:
        QMessageBox.warning(self, "Warning", "No diff provided.")
        set_status(self, "No diff provided.", "warn")
        return

    try:
        targets = _gather_checked(self)
    except Exception as e:
        QMessageBox.critical(self, "Error", f"Failed to read selection: {e}")
        set_status(self, f"Error: {e}", "error")
        return

    if not targets:
        QMessageBox.information(self, "Nothing to do", "No rows are checked for Apply.")
        set_status(self, "No rows checked.", "warn")
        return

    # Pre-parse once for speed
    diff_lines = diff_text.splitlines()

    changed, skipped, failed = 0, 0, 0
    for t in targets:
        path = t["path"]
        ow   = t["overwrite"]

        if not os.path.isfile(path):
            append_log(self, f"Skip (not a file): {path}\n")
            skipped += 1
            continue

        try:
            with open(path, "r", encoding="utf-8") as f:
                original = f.read().splitlines()

            # Recompute patch for THIS file (not using the preview text area)
            patched_text = apply_custom_diff(original, diff_lines)

            # If nothing changed, count as skipped to avoid churn
            original_text = "\n".join(original)
            if patched_text == original_text or patched_text + "\n" == original_text:
                append_log(self, f"No changes needed: {path}\n")
                skipped += 1
                continue

            out_path = path if ow else f"{path}.new"

            # Optional safety backup
            if make_backup and ow and os.path.exists(path):
                try:
                    shutil.copy2(path, path + backup_ext)
                    append_log(self, f"Backup written: {path + backup_ext}\n")
                except Exception as e:
                    append_log(self, f"Backup failed ({path}): {e}\n")

            with open(out_path, "w", encoding="utf-8") as f:
                f.write(patched_text if patched_text.endswith("\n") else patched_text + "\n")

            append_log(self, f"Saved: {out_path}\n")
            changed += 1

        except ValueError as e:
            # apply_custom_diff may raise on overlapping hunks, etc.
            append_log(self, f"Patch error for {path}: {e}\n")
            failed += 1
        except Exception as e:
            append_log(self, f"Unexpected error for {path}: {e}\n")
            failed += 1

    # User feedback
    msg = f"Saved: {changed}, Skipped: {skipped}, Failed: {failed}"
    set_status(self, msg, "ok" if changed and not failed else ("warn" if changed else "error"))
    QMessageBox.information(self, "Apply finished", msg)

def apply_diff_to_directory(self, diff_text: str) -> ApplyReport:
    report = ApplyReport()
    try:
        files = get_files(self)
    except Exception as e:
        QMessageBox.critical(self, "Error", str(e))
        set_status(self, f"Error: {e}", "error")
        return report

    if not diff_text or not diff_text.strip():
        QMessageBox.critical(self, "Error", "No diff provided.")
        set_status(self, "Error: No diff provided.", "error")
        return report

    hunks = parse_unified_diff(diff_text)
    if not hunks:
        QMessageBox.warning(self, "Warning", "No valid hunks found in diff.")
        set_status(self, "No valid hunks found.", "warn")
        return report

    file_to_replacements: dict[str, list] = defaultdict(list)

    for h in hunks:
        if not h.subs:
            report.hunks_skipped += 1
            append_log(self, "Skipping hunk with empty subs\n")
            continue

        nu_files, found_paths = getPaths(files, h.subs)
        h.content = found_paths
        any_applied = False

        for fp in found_paths:
            if not fp.get("lines"):
                continue
            start_line = fp["lines"][0]["line"]
            file_path = fp["file_path"]
            file_to_replacements[file_path].append({
                "start": start_line,
                "end": start_line + len(h.subs),
                "adds": h.adds.copy(),
                "subs": h.subs.copy()
            })
            any_applied = True

        if any_applied:
            report.hunks_applied += 1
            append_log(self, f"Applied hunk to {len(nu_files)} file(s)\n")
        else:
            report.hunks_skipped += 1
            append_log(self, "No matches found for hunk\n")

    for file_path, repls in file_to_replacements.items():
        repls_sorted = sorted(repls, key=lambda r: r["start"])
        overlaps = any(repls_sorted[i-1]["end"] > repls_sorted[i]["start"] for i in range(1, len(repls_sorted)))
        if overlaps:
            append_log(self, f"Error: Overlapping hunks in {file_path}. Skipped.\n")
            report.extend_skipped(file_path)
            continue

        try:
            og = read_any_file(file_path)
            lines = og.split("\n")
            for r in reversed(repls_sorted):
                if r["start"] >= len(lines) or r["end"] > len(lines):
                    append_log(self, f"Warning: Invalid line range in {file_path}, skipping hunk\n")
                    continue
                if lines[r["start"]:r["end"]] != r["subs"]:
                    append_log(self, f"Warning: Mismatch in {file_path}, skipping hunk\n")
                    continue
                lines = lines[:r["start"]] + r["adds"] + lines[r["end"]:]
            new_content = "\n".join(lines)
            if new_content != og and new_content + "\n" != og:
                write_to_file(new_content, f"{file_path}.new")
                report.extend_changed(file_path)
                append_log(self, f"Patched {file_path}.new\n")
            else:
                report.extend_skipped(file_path)
                append_log(self, f"No changes needed for {file_path}\n")
        except Exception as e:
            append_log(self, f"Error applying to {file_path}: {e}\n")
            report.extend_skipped(file_path)

    set_status(self, f"Applied {report.hunks_applied} hunks, skipped {report.hunks_skipped} hunks",
               "ok" if report.hunks_applied else "warn")
    return report



from PyQt6.QtWidgets import QWidget, QScrollArea, QVBoxLayout
from PyQt6.QtCore import QThreadPool

from .flow_layout import FlowLayout
from .article_card import ArticleCard
from ..workers.page_loader import PageLoader


class ArticleBrowser(QWidget):
    def __init__(self):
        super().__init__()
        self.thread_pool = QThreadPool.globalInstance()

        self.scroll = QScrollArea()
        self.scroll.setWidgetResizable(True)

        self.container = QWidget()
        self.layout = FlowLayout(self.container)
        self.scroll.setWidget(self.container)

        root = QVBoxLayout(self)
        root.addWidget(self.scroll)

        self._pending_clear = False

    def clear(self):
        while self.layout.count():
            item = self.layout.takeAt(0)
            if item:
                item.widget().deleteLater()

    def load_pages(self, start: int, end: int, user: str | None = None, skip_downloaded: bool = False):
        self._pending_clear = True

        for page in range(start, end + 1):
            worker = PageLoader(page, user, skip_downloaded=skip_downloaded)
            worker.signals.page_loaded.connect(self._add_articles)
            worker.signals.error.connect(self._on_error)
            self.thread_pool.start(worker)

    def _add_articles(self, articles):
        if self._pending_clear:
            self.clear()
            self._pending_clear = False
        for article in articles:
            self.layout.addWidget(ArticleCard(article))

    def _on_error(self, msg: str):
        print(f"[page load error] {msg}")



from .imports import *
def _editor_clear_highlights(self):
    try:
        self._hi_ranges = []
        self.editor.setExtraSelections([])  # replaces any previous overlays
    except Exception:
        pass

def _make_line_selection(self, line: int, bg: QColor):
    # IMPORTANT: use QTextEdit.ExtraSelection even with QPlainTextEdit
    sel = QTextEdit.ExtraSelection()
    tc = self.editor.textCursor()
    blk = self.editor.document().findBlockByNumber(max(1, int(line)) - 1)
    tc.setPosition(blk.position())
    tc.movePosition(QTextCursor.MoveOperation.EndOfBlock,
                    QTextCursor.MoveMode.KeepAnchor)
    sel.cursor = tc
    fmt = QTextCharFormat()
    fmt.setBackground(bg)
    sel.format = fmt
    return sel

def _editor_goto_and_mark(self, line: int, col: int | None):
    try:
        line = max(1, int(line)); col = max(1, int(col or 1))
        doc = self.editor.document()
        blk = doc.findBlockByNumber(line - 1)
        pos = blk.position() + (col - 1)

        tc = self.editor.textCursor()
        tc.setPosition(pos)
        self.editor.setTextCursor(tc)
        self.editor.centerCursor()

        focused = self._make_line_selection(line, QColor(255, 255, 0, 90))
        self.editor.setExtraSelections([focused])
    except Exception as e:
        self.append_log(f"[editor] mark error: {e}\n")

def _editor_show_ranges(self, path: str, ranges: list[dict], center: tuple[int,int]|None=None):
    try:
        if path != getattr(self, "current_file_path", None):
            first = ranges[0] if ranges else {"line": 1, "col": 1}
            self._editor_open_file(path, int(first["line"]), int(first["col"]))

        extras = []
        for r in ranges:
            ln = max(1, int(r.get("line", 1)))
            extras.append(self._make_line_selection(ln, QColor(255, 200, 0, 70)))

        if center:
            extras.append(self._make_line_selection(center[0], QColor(255, 255, 0, 110)))
            # move caret to the focused position for scrolling/column
            self._editor_goto_and_mark(center[0], center[1])

        self.editor.setExtraSelections(extras)
    except Exception as e:
        self.append_log(f"[editor] ranges error: {e}\n")



from ..imports import *
def start_collect(self):
    self.list.clear()
    self.btn_run.setEnabled(False)
    try:
        self.params = make_params(self)
    except Exception as e:
        QMessageBox.critical(self, "Bad input", str(e))
        self.btn_run.setEnabled(True)
        return
    class CollectWorker(QThread):
        log = pyqtSignal(str)
        done = pyqtSignal(list)
        def __init__(self, params):
            super().__init__()
            self.params = params
        def run(self):
            try:
                results = collect_filepaths(**self.params)
                self.done.emit(results)
            except Exception:
                self.log.emit(traceback.format_exc())
                self.done.emit([])
    self.worker = CollectWorker(self.params)
    self.worker.log.connect(self.append_log)
    self.worker.done.connect(self.populate_results)
    self.worker.finished.connect(lambda: self.btn_run.setEnabled(True))
    self.worker.start()
def append_log(self, text: str):
    self.log.moveCursor(self.log.textCursor().MoveOperation.End)
    self.log.insertPlainText(text)
def populate_results(self, results: list):
    self._last_results = results or []
    if not results:
        self.append_log("✅ No files found.\n")
        self.btn_secondary.setEnabled(False)
        return
    self.append_log(f"✅ Found {len(results)} file(s).\n")
    self.btn_secondary.setEnabled(True)
    for file_path in results:
        if isinstance(file_path, str):
            self.list.addItem(QListWidgetItem(file_path))
            self.append_log(file_path + "\n")




# abstract_x11/platform_utils.py
from ..src import *
def is_wayland() -> bool:
    plat = os.environ.get("QT_QPA_PLATFORM", "").lower()
    if not plat and os.environ.get("XDG_SESSION_TYPE") == "wayland":
        plat = "wayland"
    return plat.startswith("wayland")

def compute_self_ids(widget) -> tuple[Optional[str], Optional[str]]:
    """
    Returns (pid, win_hex).
    win_hex is None on Wayland.
    """
    pid = str(os.getpid())
    win_hex = None

    if is_wayland():
        return pid, win_hex

    try:
        wid = int(widget.winId() or 0)
        if wid:
            win_hex = f"0x{wid:08x}"
    except Exception:
        pass

    return pid, win_hex



from .build_ui import (_build_ui)
from .init_yabs_creator import (get_tabs_attr, init_functions_button_tab, init_variables_button_tab)
from .variable_filter_utils import (_on_variable_clicked, _render_var_lists_for, _on_var_filter_mode_changed, _add_var_button, _clear_var_buttons, _rebuild_var_buttons, _filter_var_buttons)
from .log_utils import (appendLog, append_log)
from .filter_utils import (_render_symbol_lists_for, _on_path_changed, _on_symbol_clicked, _on_filter_mode_changed, _start_func_scan, _on_map_ready, create_radio_group)
from .function_utils import (_on_function_clicked, _render_fn_lists_for, _on_fn_filter_mode_changed, _add_fn_button, _clear_fn_buttons, _rebuild_fn_buttons, _filter_fn_buttons, expandingDirections)



from PyQt6.QtWidgets import QFileDialog,QListWidgetItem
from abstract_utilities import define_defaults
from .resolve_directory import resolve_directory_input
import os,sys
# — UI helpers —
def browse_dir(self):
    d = QFileDialog.getExistingDirectory(
        self,
        "Choose directory",
        self.dir_in.text() or os.getcwd(),
    )
    if not d:
        return

    try:
        d = resolve_directory_input(d)
    except Exception:
        return  # or show QMessageBox

    self.dir_in.setText(d)
def on_dir_edit_finished(self):
    try:
        resolved = resolve_directory_input(self.dir_in.text())
        self.dir_in.setText(resolved)
    except Exception:
        pass  # let validation handle it later
def make_params(self):
    directory = resolve_directory_input(self.dir_in.text())

    # strings
    s_raw = [s.strip() for s in self.strings_in.text().split(",") if s.strip()]

    # allowed_exts
    e_raw = self.allowed_exts_in.text().strip()
    allowed_exts: Union[bool, Set[str]] = None
    if e_raw:
        splitter = "|" if "|" in e_raw else ","
        allowed_exts = {
            "." + e if not e.startswith(".") else e
            for e in (x.strip() for x in e_raw.split(splitter))
            if e
        }

    # exclude_exts
    ee_raw = self.exclude_exts_in.text().strip()
    exclude_exts: Union[bool, Set[str]] = None
    if ee_raw:
        splitter = "|" if "|" in ee_raw else ","
        exclude_exts = {
            "." + e if not e.startswith(".") else e
            for e in (x.strip() for x in ee_raw.split(splitter))
            if e
        }

    # allowed_types
    at_raw = self.allowed_types_in.text().strip()
    allowed_types = {e.strip() for e in at_raw.split(",") if e.strip()} if at_raw else None

    # exclude_types
    et_raw = self.exclude_types_in.text().strip()
    exclude_types = {e.strip() for e in et_raw.split(",") if e.strip()} if et_raw else None

    # allowed_dirs
    ad_raw = self.allowed_dirs_in.text().strip()
    allowed_dirs = [e.strip() for e in ad_raw.split(",") if e.strip()] if ad_raw else None

    # exclude_dirs
    ed_raw = self.exclude_dirs_in.text().strip()
    exclude_dirs = [e.strip() for e in ed_raw.split(",") if e.strip()] if ed_raw else None

    # allowed_patterns
    ap_raw = self.allowed_patterns_in.text().strip()
    allowed_patterns = [e.strip() for e in ap_raw.split(",") if e.strip()] if ap_raw else None

    # exclude_patterns
    ep_raw = self.exclude_patterns_in.text().strip()
    exclude_patterns = [e.strip() for e in ep_raw.split(",") if e.strip()] if ep_raw else None

    add = self.chk_add.isChecked()

    spec_line_val = self.spec_spin.value()
    spec_line = False if spec_line_val == 0 else int(spec_line_val)

    cfg = define_defaults(
        allowed_exts=allowed_exts,
        exclude_exts=exclude_exts,
        allowed_types=allowed_types,
        exclude_types=exclude_types,
        allowed_dirs=allowed_dirs,
        exclude_dirs=exclude_dirs,
        allowed_patterns=allowed_patterns,
        exclude_patterns=exclude_patterns,
        add=add,
    )

    return {
        "directory": directory,
        "get_lines": self.chk_getlines.isChecked(),
        "spec_line": spec_line,
        "parse_lines": self.chk_parse.isChecked(),
        "total_strings": self.chk_total.isChecked(),
        "strings": s_raw,
        "recursive": self.chk_recursive.isChecked(),
        "cfg": cfg,
    }



#!/usr/bin/env python3
"""
TDD Manifest Viewer - PyQt5 GUI
Interactive desktop application for viewing and managing image manifests.
"""
from .imports import *
from .diffParserTab import diffParserTab
from .directoryMapTab import directoryMapTab
from .extractImportsTab import extractImportsTab
from .finderTab import finderTab
from .collectFilesTab import collectFilesTab
from .scaffolder import Scaffolder

# ═══════════════════════════════════════════════════════════
# CONFIGURATION
# ═══════════════════════════════════════════════════════════

@dataclass
class AppConfig:
    """Application configuration"""
    root: str = "/srv/media/thedailydialectics"
    manifest_dir: str = "/srv/media/thedailydialectics/manifests"
    manifest_file: str = "complete_manifest.json"
    auto_refresh: bool = False
    show_missing_only: bool = False
    show_without_metadata_only: bool = False
    
    def manifest_path(self) -> str:
        return os.path.join(self.manifest_dir, self.manifest_file)


# ═══════════════════════════════════════════════════════════
# DATA MODELS
# ═══════════════════════════════════════════════════════════

@dataclass
class DirectoryStatus:
    """Status of a single image directory"""
    canonical_path: str
    filesystem_path: Optional[str]
    exists: bool
    has_info_json: bool
    has_index_html: bool
    has_resized_webp: bool
    total_files: int
    file_list: List[str]
    has_complete_metadata: bool
    missing_metadata_fields: List[str]
    used_by_pages: List[str]
    usage_count: int
    title: Optional[str]
    alt: Optional[str]
    keywords: Optional[List[str]]
    dimensions: Optional[Dict[str, int]]
    
    def status_color(self) -> QColor:
        """Get status color for UI"""
        if not self.exists:
            return QColor(220, 100, 100)  # Red
        elif self.has_complete_metadata:
            return QColor(100, 220, 100)  # Green
        else:
            return QColor(220, 200, 100)  # Yellow
    
    def status_icon(self) -> str:
        """Get status icon"""
        if not self.exists:
            return "❌"
        elif self.has_complete_metadata:
            return "✅"
        else:
            return "⚠️"
    
    def status_text(self) -> str:
        """Get status description"""
        if not self.exists:
            return "Missing"
        elif self.has_complete_metadata:
            return "Complete"
        else:
            return f"Incomplete ({len(self.missing_metadata_fields)} fields missing)"


@dataclass
class ManifestStats:
    """Statistics for the manifest"""
    total_directories: int = 0
    existing_directories: int = 0
    missing_directories: int = 0
    with_info_json: int = 0
    without_info_json: int = 0
    with_index_html: int = 0
    with_complete_metadata: int = 0
    shared_directories: int = 0


# ═══════════════════════════════════════════════════════════
# MANIFEST LOADER (Background Thread)
# ═══════════════════════════════════════════════════════════

class ManifestLoader(QThread):
    """Background thread for loading manifest"""
    
    progress = pyqtSignal(int, str)
    finished = pyqtSignal(dict, dict, object)
    error = pyqtSignal(str)
    
    def __init__(self, config: AppConfig):
        super().__init__()
        self.config = config
    
    def run(self):
        """Load manifest in background"""
        try:
            # Load manifest file
            self.progress.emit(10, "Loading manifest file...")
            manifest_path = self.config.manifest_path()
            
            if not os.path.exists(manifest_path):
                self.error.emit(f"Manifest not found: {manifest_path}")
                return
            
            with open(manifest_path, 'r') as f:
                manifest = json.load(f)
            
            self.progress.emit(30, "Building directory statuses...")
            statuses = self._build_statuses(manifest)
            
            self.progress.emit(70, "Calculating statistics...")
            stats = self._calculate_stats(statuses)
            
            self.progress.emit(100, "Complete!")
            self.finished.emit(manifest, statuses, stats)
            
        except Exception as e:
            self.error.emit(f"Error loading manifest: {e}")
    
    def _build_statuses(self, manifest: dict) -> Dict[str, DirectoryStatus]:
        """Build directory status objects"""
        statuses = {}
        path_usage = manifest.get('path_usage', {})
        
        for page in manifest.get('pages', []):
            for ref in page.get('image_refs', []):
                canonical = ref.get('canonical_path')
                if not canonical or canonical in statuses:
                    continue
                
                metadata = ref.get('metadata', {})
                dir_contents = ref.get('directory_contents', {})
                
                missing_fields = []
                if metadata:
                    if not metadata.get('alt'):
                        missing_fields.append('alt')
                    if not metadata.get('title'):
                        missing_fields.append('title')
                    if not metadata.get('keywords'):
                        missing_fields.append('keywords')
                    if not metadata.get('dimensions'):
                        missing_fields.append('dimensions')
                
                file_list = []
                if dir_contents:
                    file_list = (
                        dir_contents.get('html_files', []) +
                        dir_contents.get('json_files', []) +
                        dir_contents.get('image_files', []) +
                        dir_contents.get('other_files', [])
                    )
                
                status = DirectoryStatus(
                    canonical_path=canonical,
                    filesystem_path=ref.get('resolved_dir'),
                    exists=ref.get('exists', False),
                    has_info_json=metadata.get('has_info_json', False) if metadata else False,
                    has_index_html=dir_contents.get('has_index_html', False) if dir_contents else False,
                    has_resized_webp=dir_contents.get('has_resized_webp', False) if dir_contents else False,
                    total_files=dir_contents.get('total_files', 0) if dir_contents else 0,
                    file_list=sorted(file_list),
                    has_complete_metadata=len(missing_fields) == 0 and metadata and metadata.get('has_info_json'),
                    missing_metadata_fields=missing_fields,
                    used_by_pages=path_usage.get(canonical, []),
                    usage_count=len(path_usage.get(canonical, [])),
                    title=metadata.get('title') if metadata else None,
                    alt=metadata.get('alt') if metadata else None,
                    keywords=metadata.get('keywords') if metadata else None,
                    dimensions=metadata.get('dimensions') if metadata else None,
                )
                
                statuses[canonical] = status
        
        return statuses
    
    def _calculate_stats(self, statuses: Dict[str, DirectoryStatus]) -> ManifestStats:
        """Calculate statistics"""
        return ManifestStats(
            total_directories=len(statuses),
            existing_directories=sum(1 for s in statuses.values() if s.exists),
            missing_directories=sum(1 for s in statuses.values() if not s.exists),
            with_info_json=sum(1 for s in statuses.values() if s.has_info_json),
            without_info_json=sum(1 for s in statuses.values() if not s.has_info_json),
            with_index_html=sum(1 for s in statuses.values() if s.has_index_html),
            with_complete_metadata=sum(1 for s in statuses.values() if s.has_complete_metadata),
            shared_directories=sum(1 for s in statuses.values() if s.usage_count > 1),
        )


# ═══════════════════════════════════════════════════════════
# MAIN WINDOW
# ═══════════════════════════════════════════════════════════

class finderConsole(ConsoleBase):
    def __init__(self, *, bus=None, parent=None):
        super().__init__(bus=bus, parent=parent)
        inner = QTabWidget()
        self.layout().addWidget(inner)
    
        # all content tabs share THIS console’s bus
        inner.addTab(finderTab(self.bus),         "Find Content")
        inner.addTab(directoryMapTab(self.bus),   "Directory Map")
        #inner.addTab(extractImportsTab(self.bus), "Extract Python Imports")
        inner.addTab(diffParserTab(self.bus),     "Diff (Repo)")
        inner.addTab(Scaffolder(self.bus),     "scaffolder")

    """Main application window"""
    
    def __init__(self):
        super().__init__()
        
        # Configuration
        self.config = AppConfig()
        
        # Data
        self.manifest = {}
        self.statuses: Dict[str, DirectoryStatus] = {}
        self.stats = ManifestStats()
        self.current_filter = "all"
        
        # UI
        self.init_ui()
        
        # Auto-load manifest
        self.load_manifest()
    
    def init_ui(self):
        """Initialize user interface"""
        self.setWindowTitle("TDD Manifest Viewer")
        self.setGeometry(100, 100, 1400, 900)
        
        # Central widget
        central = QWidget()
        self.setCentralWidget(central)
        layout = QVBoxLayout(central)
        
        # Menu bar
        self.create_menu_bar()
        
        # Toolbar
        toolbar = self.create_toolbar()
        layout.addLayout(toolbar)
        
        # Main content (splitter)
        splitter = QSplitter(Qt.Horizontal)
        
        # Left: Tree view
        tree_widget = self.create_tree_view()
        splitter.addWidget(tree_widget)
        
        # Right: Details panel
        details_widget = self.create_details_panel()
        splitter.addWidget(details_widget)
        
        splitter.setSizes([600, 800])
        layout.addWidget(splitter)
        
        # Status bar
        self.statusBar = QStatusBar()
        self.setStatusBar(self.statusBar)
        self.progress_bar = QProgressBar()
        self.progress_bar.setVisible(False)
        self.statusBar.addPermanentWidget(self.progress_bar)
    
    def create_menu_bar(self):
        """Create menu bar"""
        menubar = self.menuBar()
        
        # File menu
        file_menu = menubar.addMenu("File")
        
        open_action = QAction("Open Manifest...", self)
        open_action.triggered.connect(self.open_manifest_dialog)
        file_menu.addAction(open_action)
        
        refresh_action = QAction("Refresh", self)
        refresh_action.setShortcut("F5")
        refresh_action.triggered.connect(self.load_manifest)
        file_menu.addAction(refresh_action)
        
        file_menu.addSeparator()
        
        exit_action = QAction("Exit", self)
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)
        
        # View menu
        view_menu = menubar.addMenu("View")
        
        expand_action = QAction("Expand All", self)
        expand_action.triggered.connect(lambda: self.tree.expandAll())
        view_menu.addAction(expand_action)
        
        collapse_action = QAction("Collapse All", self)
        collapse_action.triggered.connect(lambda: self.tree.collapseAll())
        view_menu.addAction(collapse_action)
        
        # Tools menu
        tools_menu = menubar.addMenu("Tools")
        
        stats_action = QAction("Show Statistics", self)
        stats_action.triggered.connect(self.show_statistics_dialog)
        tools_menu.addAction(stats_action)
    
    def create_toolbar(self) -> QHBoxLayout:
        """Create toolbar"""
        toolbar = QHBoxLayout()
        
        # Search
        toolbar.addWidget(QLabel("Search:"))
        self.search_input = QLineEdit()
        self.search_input.setPlaceholderText("Enter path or keyword...")
        self.search_input.textChanged.connect(self.filter_tree)
        toolbar.addWidget(self.search_input)
        
        # Filter
        toolbar.addWidget(QLabel("Filter:"))
        self.filter_combo = QComboBox()
        self.filter_combo.addItems([
            "All Directories",
            "Missing Only",
            "Without Metadata",
            "Without info.json",
            "Shared (Multi-use)",
            "Complete Only"
        ])
        self.filter_combo.currentTextChanged.connect(self.apply_filter)
        toolbar.addWidget(self.filter_combo)
        
        # Refresh button
        refresh_btn = QPushButton("Refresh")
        refresh_btn.clicked.connect(self.load_manifest)
        toolbar.addWidget(refresh_btn)
        
        toolbar.addStretch()
        
        return toolbar
    
    def create_tree_view(self) -> QWidget:
        """Create tree view widget"""
        container = QWidget()
        layout = QVBoxLayout(container)
        
        # Stats summary
        self.stats_label = QLabel()
        self.update_stats_label()
        layout.addWidget(self.stats_label)
        
        # Tree
        self.tree = QTreeWidget()
        self.tree.setHeaderLabels(["Directory", "Status", "Files", "Usage"])
        self.tree.setColumnWidth(0, 400)
        self.tree.itemSelectionChanged.connect(self.on_selection_changed)
        layout.addWidget(self.tree)
        
        return container
    
    def create_details_panel(self) -> QWidget:
        """Create details panel"""
        container = QWidget()
        layout = QVBoxLayout(container)
        
        # Title
        self.details_title = QLabel("Select a directory to view details")
        self.details_title.setFont(QFont("Arial", 14, QFont.Bold))
        layout.addWidget(self.details_title)
        
        # Tabs
        self.tabs = QTabWidget()
        
        # Overview tab
        overview_tab = QWidget()
        overview_layout = QVBoxLayout(overview_tab)
        self.overview_text = QTextEdit()
        self.overview_text.setReadOnly(True)
        overview_layout.addWidget(self.overview_text)
        self.tabs.addTab(overview_tab, "Overview")
        
        # Files tab
        files_tab = QWidget()
        files_layout = QVBoxLayout(files_tab)
        self.files_table = QTableWidget()
        self.files_table.setColumnCount(2)
        self.files_table.setHorizontalHeaderLabels(["File", "Type"])
        self.files_table.horizontalHeader().setStretchLastSection(True)
        files_layout.addWidget(self.files_table)
        self.tabs.addTab(files_tab, "Files")
        
        # Metadata tab
        metadata_tab = QWidget()
        metadata_layout = QVBoxLayout(metadata_tab)
        self.metadata_text = QTextEdit()
        self.metadata_text.setReadOnly(True)
        metadata_layout.addWidget(self.metadata_text)
        self.tabs.addTab(metadata_tab, "Metadata")
        
        # Usage tab
        usage_tab = QWidget()
        usage_layout = QVBoxLayout(usage_tab)
        self.usage_table = QTableWidget()
        self.usage_table.setColumnCount(1)
        self.usage_table.setHorizontalHeaderLabels(["Page Path"])
        self.usage_table.horizontalHeader().setStretchLastSection(True)
        usage_layout.addWidget(self.usage_table)
        self.tabs.addTab(usage_tab, "Usage")
        
        layout.addWidget(self.tabs)
        
        # Actions
        actions_group = QGroupBox("Actions")
        actions_layout = QHBoxLayout(actions_group)
        
        self.open_dir_btn = QPushButton("Open Directory")
        self.open_dir_btn.clicked.connect(self.open_directory)
        self.open_dir_btn.setEnabled(False)
        actions_layout.addWidget(self.open_dir_btn)
        
        self.create_info_btn = QPushButton("Create info.json")
        self.create_info_btn.clicked.connect(self.create_info_json)
        self.create_info_btn.setEnabled(False)
        actions_layout.addWidget(self.create_info_btn)
        
        actions_layout.addStretch()
        
        layout.addWidget(actions_group)
        
        return container
    
    def update_stats_label(self):
        """Update statistics label"""
        text = f"""
        <b>Statistics:</b> 
        Total: {self.stats.total_directories} | 
        Existing: {self.stats.existing_directories} | 
        Missing: {self.stats.missing_directories} | 
        With info.json: {self.stats.with_info_json} | 
        Complete: {self.stats.with_complete_metadata}
        """
        self.stats_label.setText(text)
    
    def load_manifest(self):
        """Load manifest in background"""
        self.statusBar.showMessage("Loading manifest...")
        self.progress_bar.setVisible(True)
        self.progress_bar.setValue(0)
        
        # Create loader thread
        self.loader = ManifestLoader(self.config)
        self.loader.progress.connect(self.on_load_progress)
        self.loader.finished.connect(self.on_load_finished)
        self.loader.error.connect(self.on_load_error)
        self.loader.start()
    
    def on_load_progress(self, value: int, message: str):
        """Handle load progress"""
        self.progress_bar.setValue(value)
        self.statusBar.showMessage(message)
    
    def on_load_finished(self, manifest: dict, statuses: Dict[str, DirectoryStatus], stats: ManifestStats):
        """Handle load completion"""
        self.manifest = manifest
        self.statuses = statuses
        self.stats = stats
        
        self.progress_bar.setVisible(False)
        self.statusBar.showMessage(f"Loaded {len(statuses)} directories", 3000)
        
        self.update_stats_label()
        self.populate_tree()
    
    def on_load_error(self, error: str):
        """Handle load error"""
        self.progress_bar.setVisible(False)
        self.statusBar.showMessage("Error loading manifest", 3000)
        QMessageBox.critical(self, "Error", error)
    
    def populate_tree(self):
        """Populate tree view"""
        self.tree.clear()
        
        # Build tree structure
        root_items = {}  # path -> QTreeWidgetItem
        
        for canonical, status in sorted(self.statuses.items()):
            # Apply filters
            if not self.matches_filter(status):
                continue
            
            # Parse path
            parts = [p for p in canonical.split('/') if p]
            if not parts:
                continue
            
            # Build tree path
            current_path = ""
            parent = None
            
            for i, part in enumerate(parts):
                current_path = current_path + "/" + part if current_path else "/" + part
                
                # Create item if doesn't exist
                if current_path not in root_items:
                    if parent is None:
                        item = QTreeWidgetItem(self.tree)
                    else:
                        item = QTreeWidgetItem(parent)
                    
                    item.setText(0, part)
                    item.setData(0, Qt.UserRole, current_path)
                    root_items[current_path] = item
                    
                    # If this is the final directory, set status
                    if i == len(parts) - 1:
                        item.setText(1, status.status_icon() + " " + status.status_text())
                        item.setText(2, str(status.total_files))
                        item.setText(3, str(status.usage_count) + (" pages" if status.usage_count > 1 else " page"))
                        item.setBackground(1, status.status_color())
                        item.setData(0, Qt.UserRole + 1, canonical)  # Store canonical path
                
                parent = root_items[current_path]
    
    def matches_filter(self, status: DirectoryStatus) -> bool:
        """Check if status matches current filter"""
        filter_text = self.filter_combo.currentText()
        
        if filter_text == "All Directories":
            return True
        elif filter_text == "Missing Only":
            return not status.exists
        elif filter_text == "Without Metadata":
            return not status.has_complete_metadata
        elif filter_text == "Without info.json":
            return not status.has_info_json
        elif filter_text == "Shared (Multi-use)":
            return status.usage_count > 1
        elif filter_text == "Complete Only":
            return status.has_complete_metadata
        
        return True
    
    def filter_tree(self, text: str):
        """Filter tree by search text"""
        if not text:
            # Show all
            self.populate_tree()
            return
        
        # Hide all items
        iterator = QTreeWidgetItemIterator(self.tree)
        while iterator.value():
            item = iterator.value()
            item.setHidden(True)
            iterator += 1
        
        # Show matching items and their parents
        text_lower = text.lower()
        iterator = QTreeWidgetItemIterator(self.tree)
        while iterator.value():
            item = iterator.value()
            if text_lower in item.text(0).lower():
                # Show this item and all parents
                current = item
                while current:
                    current.setHidden(False)
                    current.setExpanded(True)
                    current = current.parent()
            iterator += 1
    
    def apply_filter(self, filter_text: str):
        """Apply filter"""
        self.populate_tree()
    
    def on_selection_changed(self):
        """Handle tree selection change"""
        items = self.tree.selectedItems()
        if not items:
            return
        
        item = items[0]
        canonical = item.data(0, Qt.UserRole + 1)
        
        if canonical and canonical in self.statuses:
            self.show_details(self.statuses[canonical])
        else:
            self.clear_details()
    
    def show_details(self, status: DirectoryStatus):
        """Show directory details"""
        self.details_title.setText(status.canonical_path)
        
        # Overview
        overview = f"""
<h3>Status: {status.status_icon()} {status.status_text()}</h3>

<p><b>Canonical Path:</b> {status.canonical_path}</p>
<p><b>Filesystem Path:</b> {status.filesystem_path or 'N/A'}</p>
<p><b>Exists:</b> {'Yes' if status.exists else 'No'}</p>
<p><b>Total Files:</b> {status.total_files}</p>

<h4>Features:</h4>
<ul>
<li>{'✅' if status.has_info_json else '❌'} info.json</li>
<li>{'✅' if status.has_index_html else '❌'} index.html</li>
<li>{'✅' if status.has_resized_webp else '❌'} resized.webp</li>
</ul>

<h4>Usage:</h4>
<p>Used by <b>{status.usage_count}</b> page(s)</p>
"""
        
        if status.missing_metadata_fields:
            overview += f"""
<h4 style="color: #cc6600;">Missing Metadata Fields:</h4>
<p>{', '.join(status.missing_metadata_fields)}</p>
"""
        
        self.overview_text.setHtml(overview)
        
        # Files table
        self.files_table.setRowCount(len(status.file_list))
        for i, filename in enumerate(status.file_list):
            self.files_table.setItem(i, 0, QTableWidgetItem(filename))
            ext = os.path.splitext(filename)[1]
            file_type = {
                '.html': 'HTML',
                '.json': 'JSON',
                '.webp': 'Image (WebP)',
                '.jpg': 'Image (JPEG)',
                '.png': 'Image (PNG)',
            }.get(ext, 'Other')
            self.files_table.setItem(i, 1, QTableWidgetItem(file_type))
        
        # Metadata
        metadata_html = "<h3>Metadata</h3>"
        if status.title:
            metadata_html += f"<p><b>Title:</b> {status.title}</p>"
        if status.alt:
            metadata_html += f"<p><b>Alt Text:</b> {status.alt}</p>"
        if status.keywords:
            metadata_html += f"<p><b>Keywords:</b> {', '.join(status.keywords)}</p>"
        if status.dimensions:
            metadata_html += f"<p><b>Dimensions:</b> {status.dimensions['width']} × {status.dimensions['height']}</p>"
        
        if not any([status.title, status.alt, status.keywords, status.dimensions]):
            metadata_html += "<p><i>No metadata available</i></p>"
        
        self.metadata_text.setHtml(metadata_html)
        
        # Usage table
        self.usage_table.setRowCount(len(status.used_by_pages))
        for i, page_path in enumerate(status.used_by_pages):
            page_rel = page_path.split('/pages/')[-1] if '/pages/' in page_path else page_path
            self.usage_table.setItem(i, 0, QTableWidgetItem(page_rel))
        
        # Enable action buttons
        self.open_dir_btn.setEnabled(status.exists)
        self.create_info_btn.setEnabled(status.exists and not status.has_info_json)
    
    def clear_details(self):
        """Clear details panel"""
        self.details_title.setText("Select a directory to view details")
        self.overview_text.clear()
        self.files_table.setRowCount(0)
        self.metadata_text.clear()
        self.usage_table.setRowCount(0)
        self.open_dir_btn.setEnabled(False)
        self.create_info_btn.setEnabled(False)
    
    def open_directory(self):
        """Open selected directory in file manager"""
        items = self.tree.selectedItems()
        if not items:
            return
        
        item = items[0]
        canonical = item.data(0, Qt.UserRole + 1)
        
        if canonical and canonical in self.statuses:
            status = self.statuses[canonical]
            if status.filesystem_path:
                os.system(f'xdg-open "{status.filesystem_path}" &')
    
    def create_info_json(self):
        """Create info.json template for selected directory"""
        items = self.tree.selectedItems()
        if not items:
            return
        
        item = items[0]
        canonical = item.data(0, Qt.UserRole + 1)
        
        if canonical and canonical in self.statuses:
            status = self.statuses[canonical]
            if status.filesystem_path and not status.has_info_json:
                info_path = os.path.join(status.filesystem_path, 'info.json')
                
                template = {
                    "alt": "",
                    "caption": "",
                    "keywords_str": "",
                    "filename": "",
                    "ext": "webp",
                    "title": "",
                    "dimensions": {"width": 0, "height": 0},
                    "file_size": "",
                    "license": "CC BY-SA 4.0",
                    "attribution": "Created by The Daily Dialectics for educational purposes"
                }
                
                with open(info_path, 'w') as f:
                    json.dump(template, f, indent=4)
                
                QMessageBox.information(self, "Success", f"Created: {info_path}")
                self.load_manifest()  # Refresh
    
    def open_manifest_dialog(self):
        """Open manifest file dialog"""
        filename, _ = QFileDialog.getOpenFileName(
            self,
            "Open Manifest",
            self.config.manifest_dir,
            "JSON Files (*.json)"
        )
        
        if filename:
            self.config.manifest_file = os.path.basename(filename)
            self.config.manifest_dir = os.path.dirname(filename)
            self.load_manifest()
    
    def show_statistics_dialog(self):
        """Show statistics dialog"""
        stats_text = f"""
<h2>Manifest Statistics</h2>

<table>
<tr><td><b>Total Directories:</b></td><td>{self.stats.total_directories}</td></tr>
<tr><td><b>Existing:</b></td><td>{self.stats.existing_directories}</td></tr>
<tr><td><b>Missing:</b></td><td>{self.stats.missing_directories}</td></tr>
<tr><td><b>With info.json:</b></td><td>{self.stats.with_info_json}</td></tr>
<tr><td><b>Without info.json:</b></td><td>{self.stats.without_info_json}</td></tr>
<tr><td><b>With index.html:</b></td><td>{self.stats.with_index_html}</td></tr>
<tr><td><b>Complete Metadata:</b></td><td>{self.stats.with_complete_metadata}</td></tr>
<tr><td><b>Shared (Multi-use):</b></td><td>{self.stats.shared_directories}</td></tr>
</table>
"""
        
        msg = QMessageBox(self)
        msg.setWindowTitle("Statistics")
        msg.setTextFormat(Qt.RichText)
        msg.setText(stats_text)
        msg.exec_()




from .imports import *

def enable_widget(parent, name: str, enabled: bool):
    try:
        getattr(parent, name).setEnabled(enabled)
    except AttributeError:
        print(f"[WARN] No widget {name} in {parent}")

class finderTab(QWidget):
    def __init__(self, bus: SharedStateBus):
        super().__init__()
        initFuncs(self)
        self.setLayout(QVBoxLayout())
        grid = QGridLayout()

        install_common_inputs(
            self, grid, bus=bus,
            primary_btn=("Run search", self.start_search),
            secondary_btn=("Stop search", self.stop_search)


        )
        # Output area
        self.layout().addWidget(QLabel("Results"))
        self.lines_list = []
        self.list = QListWidget()
        self.list.itemDoubleClicked.connect(self.open_one)
        self.layout().addWidget(self.list, stretch=3)
        self._last_results = []



from .imports import *
from .emulatorConfig import *
from .politeFetcher import *
from .baseEngine import *
from .seleneumEngine import *
class PlaywrightEngine(BaseEngine):
    def __init__(self, cfg: EmulatorConfig):
        super().__init__(cfg)
        if not PLAYWRIGHT_AVAILABLE:
            raise RuntimeError("Playwright is not installed. `pip install playwright` and `python -m playwright install chromium`.")
        self._pl = sync_playwright().start()
        self.browser: Optional[Browser] = None
        self.context = None
        self.page: Optional[Page] = None
        self._init_browser()

    @staticmethod
    def _parse_proxy(p: str) -> Dict[str, str]:
        # supports http://user:pass@host:port or http(s)://host:port
        u = urlparse(p)
        d = {"server": f"{u.scheme}://{u.hostname}:{u.port}"}
        if u.username or u.password:
            d["username"] = u.username or ""
            d["password"] = u.password or ""
        return d

    def _init_browser(self):
        launch_args = {
            "headless": self.cfg.headless,
            "args": [
                "--disable-dev-shm-usage",
                "--no-sandbox",
            ]
        }
        if self.proxy_pool:
            launch_args["proxy"] = self._parse_proxy(random.choice(self.proxy_pool))
        elif self.cfg.proxy:
            launch_args["proxy"] = self._parse_proxy(self.cfg.proxy)

        self.browser = self._pl.chromium.launch(**launch_args)
        ctx_args = {
            "user_agent": self.user_agent,
            "java_script_enabled": True,
            "viewport": {"width": 1280, "height": 800}
        }
        self.context = self.browser.new_context(**ctx_args)

        # Block heavy resources for speed
        if self.cfg.disable_images:
            def _route(route, request):
                if request.resource_type in ("image","font","media"):
                    return route.abort()
                return route.continue_()
            self.context.route("**/*", _route)

        self.page = self.context.new_page()

    def _goto(self, url: str, wait_for: Optional[str]):
        self.page.goto(url, wait_until="domcontentloaded", timeout=self.cfg.timeout * 1000)
        # brief network idle
        self.page.wait_for_timeout(600)
        if wait_for:
            self.page.wait_for_selector(wait_for, timeout=self.cfg.timeout * 1000)

    def scrape(self, url: str, wait_for: Optional[str], selectors: Dict[str, str]) -> Dict[str, Any]:
        if not self.polite.allowed(url):
            return {'url': url, 'status': 'skipped_by_robots'}
        self.polite.sleep()
        try:
            self._goto(url, wait_for)
            # light human-ish activity
            self.page.mouse.move(50, 50)
            self.page.wait_for_timeout(150)
            html = self.page.content()
            soup = BeautifulSoup(html, 'html.parser')
            title = soup.title.string.strip() if soup.title else None
            data = self._extract(soup, selectors) if selectors else {}
            cookies = {c['name']: c['value'] for c in self.context.cookies()}
            return {'url': url, 'status': 'success', 'title': title, 'data': data, 'soup': soup, 'html': html, 'cookies': cookies}
        except Exception as e:
            logger.warning(f"[Playwright] Failed {url}: {e}")
            # fallback with requests
            try:
                r = self.session.get(url, timeout=self.cfg.timeout)
                r.raise_for_status()
                soup = BeautifulSoup(r.text, 'html.parser')
                title = soup.title.string.strip() if soup.title else None
                data = self._extract(soup, selectors) if selectors else {}
                return {'url': url, 'status': 'fallback_success', 'title': title, 'data': data, 'soup': soup, 'cookies': r.cookies.get_dict()}
            except Exception as e2:
                return {'url': url, 'status': 'failed', 'error': str(e2)}

    def screenshot(self, url: str, out_path: str):
        try:
            self._goto(url, wait_for=None)
            self.page.screenshot(path=out_path, full_page=True)
        except Exception:
            pass

    def quit(self):
        try:
            if self.context: self.context.close()
        except Exception:
            pass
        try:
            if self.browser: self.browser.close()
        except Exception:
            pass
        try:
            if self._pl: self._pl.stop()
        except Exception:
            pass
        self.browser = None
        self.page = None
        self.context = None



from .imports import *
# Data structures
def enable_widget(parent, name: str, enabled: bool):
    try:
        getattr(parent, name).setEnabled(enabled)
    except AttributeError:
        print(f"[WARN] No widget {name} in {parent}")

# — Actions —
def start_search(self):
    try:
        reset_find_console_stop()  # reset flag before starting
        
        enable_widget(self, "primary_btn", False)
        enable_widget(self, "btn_secondary", True)   # enable stop button
        
        try:
            params = self.make_params(self)
        except Exception as e:
            logger.info(f"{e}")
        logger.info(f"params == {params}")
        self.worker = SearchWorker(params)
        
        self.worker.log.connect(self.append_log)
        self.worker.done.connect(self.populate_results)
        
        self.worker.finished.connect(lambda: enable_widget(self,"primary_btn",True))
        self.worker.start()
    except Exception as e:
        logger.info(e)
def stop_search(self):
    if hasattr(self, "worker") and self.worker.isRunning():
        request_find_console_stop()
        enable_widget(self, "primary_btn", True)
        enable_widget(self, "btn_secondary", False)

def append_log(self, text: str):
    """
    Append text to the tab's log widget (QPlainTextEdit or QTextEdit).
    Safe if self.log is missing.
    """
    try:
        edit = getattr(self, "log", None)

        # Prefer QPlainTextEdit (faster for logs)
        if isinstance(edit, QPlainTextEdit):
            if not text.endswith("\n"):
                text += "\n"
            edit.appendPlainText(text)
            return

        # QTextEdit fallback
        if isinstance(edit, QTextEdit):
            if not text.endswith("\n"):
                text += "\n"
            cursor = edit.textCursor()
            cursor.movePosition(QtGui.QTextCursor.MoveOperation.End)
            edit.setTextCursor(cursor)
            edit.insertPlainText(text)
            return

        # No log widget? Avoid crashing, at least surface somewhere:
        try:
            print(text, end="" if text.endswith("\n") else "\n")
        except Exception:
            pass
    except Exception as e:
        logger.info(e)
def populate_results(self, results: list):
    try:
        self._last_results = results or []
        self.list.clear()
        if not results:
            self.append_log("✅ No matches found.\n")
            enable_widget(self, "btn_secondary", False)
            return

        self.append_log(f"✅ Found {len(results)} file(s).\n")
        enable_widget(self, "btn_secondary", True)
        self.lines_list = {}
        for fp in results:
            if isinstance(fp, dict):
                file_path = fp.get("file_path")
                lines = fp.get("lines", [])
            else:
                file_path = fp
                lines = []

            if not isinstance(file_path, str):
                continue

            if lines:
                for obj in lines:
                    line = obj.get("line")
                    text = f"{file_path}" if line is not None else file_path
                    if file_path not in self.lines_list:
                        self.lines_list[file_path] = []
                        item = QListWidgetItem(text)
                        item.setData(Qt.ItemDataRole.UserRole, {"file_path": file_path, "line": line})
                        self.list.addItem(item)
                    self.lines_list[file_path].append(line)
            else:
                text = f"{file_path}" if line is not None else file_path
                if file_path not in self.lines_list:
                    item = QListWidgetItem(file_path)
                    item.setData(Qt.ItemDataRole.UserRole, {"file_path": file_path, "line": None})
                    self.list.addItem(item)
                    self.append_log(file_path + "\n")
            
    except Exception as e:
        logger.info(e)
